/*
 * EfficientSPC -- Header
 *
 * Author:  Henrik Zimmer <henrik@zimmer.to>
 *
 * Version: $Revision: 1 $
 * Date:    $date$
 */

#ifndef ACG_EFFICIENTSPC_HH
#define ACG_EFFICIENTSPC_HH

////////////////////////////////////////////////////////

#include "Types.hh"
#include "TetFMM.hh"
#include <tetgen.h>

#include <ACG/Utils/StopWatch.hh>

#include <ACG/Geometry/bsp/TriangleBSPT.hh> 

////////////////////////////////////////////////////////

#ifndef TETLIBRARY
#define TETLIBRARY 1 
#endif

namespace ACG {


  ////! Light-weight TetMesh data structure for effciently setup and handle the tetrahedrilization.
  /* See www.openvolumemesh.org for a volumetric mesh structure for general polyhedral meshes */
class LeanTetMesh
{

  public:
  struct Edge
  {
    Edge( ) : id0(-1), id1(-1) {}
    Edge( int _id0, int _id1) : id0(_id0), id1(_id1) {}
    int id0;
    int id1;
  };

  public:
    LeanTetMesh( ) {}
    LeanTetMesh( size_t _nv, size_t _nt ) :
      nv(_nv), nt(_nt), 
      tets(_nt, std::vector< int >(4,-1)), 
      pts(_nv),
      vv(_nv),
      vc(_nv),
      tag(_nv, 0),
      ve(_nv)
    {}

    size_t nv;
    size_t nt;
    size_t ne;

    std::vector< std::vector< int > > tets;
    std::vector< Vec3 > pts;
    std::vector< std::set< int > > vv;
    std::vector< std::set< int > > vc;
    std::vector< int > tag;
    std::vector< std::vector< Edge > > ve;
    std::vector< Edge > edges;

    void build_edges()
    {
      edges.clear();
      for( int i = 0; i < (int)vv.size(); ++i)
      {
        int cnt(0);
        for( std::set<int>::iterator it( vv[i].begin()); it != vv[i].end(); ++it)
        {
          if( i < *it)
            edges.push_back( Edge(i, *it));
          ve[i].resize( vv[i].size());
          ve[i][cnt] = Edge(i,*it);
          cnt++;
        }
      }
      ne = edges.size();
    }
};

////////////////////////////////////////////////////////

class EfficientSPC
{
public:
   
  /// Constructor
  EfficientSPC( MeshT& _mesh) : mesh_(_mesh) {}
 
  /// Destructor
  ~EfficientSPC() {}

  static Scalar tetrahedralize( MeshT& _mesh, PMeshT& _robusthull, Scalar _avgel, LeanTetMesh& _tetmesh, Scalar _edgescale = 5.0, Scalar _aspect = 2.4 );
  static Scalar do_convex_hull( MeshT& _inpmesh, MeshT& _outmesh) 
  { return do_convex_hull_tetgen( _inpmesh, _outmesh);} 
  static Scalar do_convex_hull_tetgen( MeshT& _inpmesh, MeshT& _outmesh);
  static Scalar do_convex_hull_cgal( MeshT& _inpmesh, MeshT& _outmesh);

  static void fmm( LeanTetMesh& _tetmesh, MeshT& _hull, OpenMeshTriangleBSPT< MeshT >*& _chbsp, Scalar _avgel, MeshT& _mesh, Scalar& _inttime, Scalar& _dijtime);

  static Vec3 closest_bsp_point( const Vec3& _pt, const MeshT& _bspmesh, OpenMeshTriangleBSPT< MeshT>* & _bsp);
  static FH closest_bsp_fh( const Vec3& _pt, OpenMeshTriangleBSPT< MeshT>* & _bsp);
  static void closest_bsp_ray( const Vec3& _pt, const Vec3& _ray, OpenMeshTriangleBSPT< MeshT>* & _bsp, std::vector< std::pair< FH, Scalar > >& _collisions);

  static Scalar tri_area( const FH& _fh, const MeshT& _mesh);
  static Scalar tri_area( const std::vector< Vec3 >& _pts);


private:
  
  MeshT& mesh_;
};


} // namespace ACG

#endif // ACG_EFFICIENTSPC_HH defined

