/*
 * TetFMM -- Header/Implementation
 *  For computing FMM on tetrahedral meshes
 *  Based on:
 *     "Computing first-arrival seismic traveltimes on unstructured 3-D tetrahedral grids using the Fast Marching Method" 
 *     by Peter G. Leli`vre, Colin G. Farquharson and Charles A. Hurich, 2010
 *
 * Author:  Henrik Zimmer <henrik@zimmer.to>
 *
 * Version: $Revision: 1 $
 * Date:    $date$
 */


#ifndef ACG_TETFMM_HH
#define ACG_TETFMM_HH


namespace ACG {


////////////////////////////////////////////////////////

class TetFMM
{
public: 

static bool fmm_tri( Scalar* _v, Vec3* _pt)
{
  return fmm_tri( _v[0], _v[1], _v[2], _pt[0], _pt[1], _pt[2]);
}

////////////////////////////////////////////////////////

static bool fmm_tri( const Scalar& _ta, const Scalar& _tb, Scalar& _tc, const Vec3& _pa, const Vec3& _pb, const Vec3& _pc)
{
  const Scalar thres(1e-9);
  const Scalar a( (_pc-_pb).norm());
  const Scalar b( (_pa-_pc).norm());
  const Scalar c( (_pa-_pb).norm());
  const Scalar s(1.0);

  _tc = std::min( _tc, std::min( _ta + s*b, _tb + s*a));

  const Scalar u( _tb-_ta);

  const Scalar w2( s*s*c*c - u*u);

  const Scalar eps0_meshmetric( ( (_pb - _pa) | (_pc-_pa) ) / (c));
  const Scalar eps0_normalized( eps0_meshmetric / c);
  const Scalar rho0( sqrt( b*b - eps0_meshmetric*eps0_meshmetric));


  if( w2 >= thres)
  {
    // compute epsilon
    Scalar eps( eps0_normalized);
    if( u > 0.)
      eps -= 1.0/sqrt(w2)*1.0/c*u*rho0;
    else
      eps += 1.0/sqrt(w2)*1.0/c*u*rho0;

    if( eps > 0. && eps < 1.0)
    {
      _tc = std::min( _ta + u*eps0_normalized+sqrt(w2)/c*rho0, _tc);
      return true;
    }
    return false;
  }

  return false;
}

////////////////////////////////////////////////////////

static bool fmm_tet( Scalar* _v, Vec3* _p)
{
  return fmm_tet( _v[0], _v[1], _v[2], _v[3], _p[0], _p[1], _p[2], _p[3]);
}

////////////////////////////////////////////////////////

static bool fmm_tet( const Scalar& _ta, const Scalar& _tb, const Scalar& _tc, Scalar& _td, 
    const Vec3& _pa, const Vec3& _pb, const Vec3& _pc, const Vec3& _pd)
{
  const Scalar thres(1e-9);
  const Scalar a( (_pc-_pb).norm());
  const Scalar b( (_pa-_pc).norm());
  const Scalar c( (_pa-_pb).norm());
  const Scalar s(1.0);

  // edge updates
  _td = std::min(_td, std::min( _ta + s*(_pd-_pa).norm(), std::min(_tb + s*(_pd-_pb).norm(), _tc + s*(_pd-_pc).norm())));

  // head waves over faces
  fmm_tri( _ta, _tb, _td, _pa, _pb, _pd);
  fmm_tri( _tc, _ta, _td, _pc, _pa, _pd);
  fmm_tri( _tb, _tc, _td, _pb, _pc, _pd);

  const Vec3 AB( _pb-_pa);
  const Vec3 AC( _pc-_pa);

  const Vec3 normal( (AB%AC).normalize());
  const Scalar rho0( (_pd-_pa)|normal);
  const Vec3 proj_meshmetric( _pd-rho0*normal);

  // least squares solution for epsilon_0 and xi_0
  const Scalar m00( AB|AB);
  const Scalar m10( AB|AC);
  const Scalar m01( m10);
  const Scalar m11( AC|AC);
  const Scalar mdetinv( 1.0/(m00*m11 - m01*m10));

  const Vec3 AP( proj_meshmetric - _pa);

  // invert M
  Scalar mi00(  m11*mdetinv);
  Scalar mi11(  m00*mdetinv);
  Scalar mi10( -m10*mdetinv);
  Scalar mi01( -m01*mdetinv);
  Scalar rhs0( AB|AP);
  Scalar rhs1( AC|AP);

  // solution
  Scalar eps0( mi00 * rhs0 + mi01 * rhs1);
  Scalar xi0( mi10 * rhs0 + mi11 * rhs1);
  //std::cerr << _pa << " " << _pb << " " << _pc << " " << _pd << std::endl;
  //std::cerr << rho0 << " " << proj_meshmetric << std::endl;
  //std::cerr << eps0 << " " << xi0 << std::endl;
  //exit(1);

  const Scalar u( _tb - _ta);
  const Scalar v( _tc - _ta);

  const Scalar alpha( acos( std::min( 1.0, std::max( -1.0, m01)))); // warning reuseing m01 (don't change it!)
  const Scalar d2( m10); // warning reuse!
  const Scalar phi( c * b * sin( alpha));


  const Scalar w2( s*s*phi*phi - u*u*b*b-v*v*c*c+2.*u*v*d2);

  if( w2 >= thres )
  {
    const Scalar beta( u * b * b - v * d2);
    const Scalar gamma( v * c * c - u * d2);
    const Scalar eta( beta/gamma);


    Scalar eps( eps0);
    Scalar xi( xi0);

    if( u > 0.)
      eps -= fabs(beta)/(phi*sqrt(w2))*rho0;
    else
      eps += fabs(beta)/(phi*sqrt(w2))*rho0;

    if( v > 0.)
      xi -= fabs(gamma)/(phi*sqrt(w2))*rho0;
    else
      xi += fabs(gamma)/(phi*sqrt(w2))*rho0;

    if( ( 0. < eps ) && ( eps < 1. ))
      if( ( 0. < xi ) && ( xi < 1. ))
        if( ( 0. < (xi + eps)) && ( (xi + eps) < 1.))
        {
          _td = std::min( _td, _ta + u * eps0 + v * xi0 + sqrt(w2) * rho0 / phi);
          return true;
        }
  }
  return false;

}

////////////////////////////////////////////////////////

};

}// namespace ACG
#endif

